#include <stdlib.h>
#include <math.h>
#include "circular.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

Circular::Circular(int w, int h) {
	width = w;
	height = h;
	sintable = NULL;
	disttable = NULL;
}

Circular::~Circular(void) {
	free(sintable);
	free(disttable);
}

void Circular::init(void) {
	int i;
	xvar = (512-width)/2;
	yvar = (512-height)/2;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (int) (128+127*sin(2*M_PI*i/512));

	disttable = (unsigned int*) malloc(sizeof(int)*512*512);
	unsigned int *ptr = disttable;
	for (int y = 0; y < 512; y++)
		for (int x = 0; x < 512; x++)
			*ptr++ = (int) ((1<<8)*hypot(x-256,y-256));

	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Circular::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	for (int i = 0; i < 256; i++)
		*ptr++ = i<<16 | i<<8;
	return pal;
}

void Circular::draw(unsigned int *buffer, int time) {
	int offsetX1 = (int) (xvar+xvar*sin((double)(1950+time)/500));
	int offsetY1 = (int) (yvar+yvar*cos((double)(4650+time)/900));
	int offsetX2 = (int) (xvar+xvar*sin((double)(6200+time)/650));
	int offsetY2 = (int) (yvar+yvar*cos((double)(1250+time)/550));

	int size = (int) ((1<<8)*(15+5*sin((double)(1200+time)/700)));

	int dist1, dist2, c1, c2;
	unsigned int *distptr1, *distptr2;
	for (int y = 0; y < height; y++) {
		distptr1 = disttable + ((y+offsetY1)<<9)+offsetX1;
		distptr2 = disttable + ((y+offsetY2)<<9)+offsetX2;
		for (int x = 0; x < width; x++) {
			dist1 = *distptr1++;
			dist2 = *distptr2++;
			c1 = sintable[((dist1*size)>>16)&0x01FF];
			c2 = sintable[((dist2*size)>>16)&0x01FF];
			if (dist1>dist2)
				*buffer++ = (c1>c2)?c1:c2;
			else
				*buffer++ = (c1<c2)?c1:c2;
		}
	}
}


