#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "cube.h"

Cube::Cube(int w, int h, unsigned int c) {
	width = w;
	height = h;
	color = c;
}

Cube::~Cube(void) {
}

void Cube::init(void) {
	v[0].x = 1;
	v[0].y = -1;
	v[0].z = 2;
	v[1].x = -1;
	v[1].y = -1;
	v[1].z = 2;
	v[2].x = -1;
	v[2].y = 1;
	v[2].z = 2;
	v[3].x = 1;
	v[3].y = 1;
	v[3].z = 2;
	v[4].x = 1;
	v[4].y = -1;
	v[4].z = 4;
	v[5].x = -1;
	v[5].y = -1;
	v[5].z = 4;
	v[6].x = -1;
	v[6].y = 1;
	v[6].z = 4;
	v[7].x = 1;
	v[7].y = 1;
	v[7].z = 4;

	rotateVector.x = 0;
	rotateVector.y = 0;
	rotateVector.z = 3;

	lastTime = 0;
}

static void rotateZ(struct vector *v, float rotate, struct vector *c) {
	float r = hypot(v->x-c->x,v->y-c->y);
	float a = atan2(v->y-c->y,v->x-c->x);
	v->x = c->x + r*cos(a+rotate);
	v->y = c->y + r*sin(a+rotate);
}

static void rotateY(struct vector *v, float rotate, struct vector *c) {
	float r = hypot(v->x-c->x,v->z-c->z);
	float a = atan2(v->z-c->z,v->x-c->x);
	v->x = c->x + r*cos(a+rotate);
	v->z = c->z + r*sin(a+rotate);
}

static void rotateX(struct vector *v, float rotate, struct vector *c) {
	float r = hypot(v->y-c->y,v->z-c->z);
	float a = atan2(v->z-c->z,v->y-c->y);
	v->y = c->y + r*cos(a+rotate);
	v->z = c->z + r*sin(a+rotate);
}

void Cube::draw(unsigned int *buffer, int time, bool clear) {
	if (clear) memset(buffer,0,width*height*sizeof(int));

	int diffTime = time-lastTime;
	lastTime = time;

	float zAngle = (float)diffTime/900;
	float yAngle = (float)diffTime/1200;
	float xAngle = (float)diffTime/1300;

	for (int i = 0; i < 8; i++) {
		rotateZ(&v[i],zAngle,&rotateVector);
		rotateY(&v[i],yAngle,&rotateVector);
		rotateX(&v[i],xAngle,&rotateVector);
	}

	drawVertexLine(&v[0],&v[1],buffer);
	drawVertexLine(&v[1],&v[2],buffer);
	drawVertexLine(&v[2],&v[3],buffer);
	drawVertexLine(&v[3],&v[0],buffer);
	drawVertexLine(&v[4],&v[5],buffer);
	drawVertexLine(&v[5],&v[6],buffer);
	drawVertexLine(&v[6],&v[7],buffer);
	drawVertexLine(&v[7],&v[4],buffer);
	drawVertexLine(&v[0],&v[4],buffer);
	drawVertexLine(&v[1],&v[5],buffer);
	drawVertexLine(&v[2],&v[6],buffer);
	drawVertexLine(&v[3],&v[7],buffer);
}

void Cube::drawVertexLine(struct vector *v1, struct vector *v2, unsigned int *buffer) {
	int x1=0,y1=0,x2=0,y2=0;

	getScreenCoord(v1,&x1,&y1);
	getScreenCoord(v2,&x2,&y2);
	drawLine(x1,y1,x2,y2,buffer);
	drawLine(x1+1,y1,x2+1,y2,buffer);
	drawLine(x1,y1+1,x2,y2+1,buffer);
	drawLine(x1+1,y1+1,x2+1,y2+1,buffer);
}

static const float screenZ = 1.5;
static const float screenY1 = -1.2;
static const float screenY2 = 1.2;
static const float screenX1 = -1.6;
static const float screenX2 = 1.6;

void Cube::getScreenCoord(struct vector *vec, int *x, int *y) {
	float yCoord = screenZ*vec->y/vec->z;
	float xCoord = screenZ*vec->x/vec->z;

	if (yCoord<screenY1 || yCoord>screenY2) return;
	if (xCoord<screenX1 || xCoord>screenX2) return;

	*y = (int) (height*(screenY2-yCoord)/(screenY2-screenY1));
	*x = (int) (width*(xCoord-screenX1)/(screenX2-screenX1));
}

void Cube::drawLine(int x1, int y1, int x2, int y2, unsigned int *buffer) {
	int deltaX = x2-x1;
	int deltaY = y2-y1;
	int deltaXAbs = (deltaX<0) ? -deltaX : deltaX;
	int deltaYAbs = (deltaY<0) ? -deltaY : deltaY;
	int signDeltaX = (deltaX<0) ? -1 : (deltaX>0) ? 1 : 0;
	int signDeltaY = (deltaY<0) ? -1 : (deltaY>0) ? 1 : 0;
	int x = deltaYAbs>>1;
	int y = deltaXAbs>>1;
	int currentX = x1;
	int currentY = y1;

	buffer[currentY*width+currentX] = color;
	if (deltaXAbs >= deltaYAbs) {
		for (int i = 0; i < deltaXAbs; i++) {
			y += deltaYAbs;
			if (y>=deltaXAbs) {
				y -= deltaXAbs;
				currentY += signDeltaY;
			}
			currentX += signDeltaX;
			buffer[currentY*width+currentX] = color;
		}
	} else {
		for (int i = 0; i < deltaYAbs; i++) {
			x += deltaXAbs;
			if (x>=deltaYAbs) {
				x -= deltaYAbs;
				currentX += signDeltaX;
			}
			currentY += signDeltaY;
			buffer[currentY*width+currentX] = color;
		}
	}
}

