#include <stdio.h>
#include <dsound.h>
#include "audio.h"

// -fvtable-thunks
// -ldsound

#define BUFFER_SIZE 1024
#define BUFFER_STEPS 20
#define SLEEP_LENGTH 15

static LPDIRECTSOUND ds;
static LPDIRECTSOUNDBUFFER soundBuffer;
static int stop;
static int playing = 0;

int channels = 2;
int samplerate = 44100;
int bitdepth = 16;
int sig = 1;
int littleendian = 1;

static HWND getHWnd(void) {
	HWND hWnd;
	hWnd = GetForegroundWindow();
	if (hWnd!=NULL) return hWnd;
	hWnd = GetDesktopWindow();
	return hWnd;
}

int init_audio(void) {
	HWND hWnd;
	WAVEFORMATEX format;
	DSBUFFERDESC bufferDesc;

	if (DirectSoundCreate(NULL,&ds,NULL)!=DS_OK) {
		printf("Can't initialize DirectSound\n");
		return 0;
	}
	hWnd = getHWnd();
	if (ds->SetCooperativeLevel(hWnd,DSSCL_PRIORITY)!=DS_OK) {
		printf("Can't set cooperative level\n");
		return 0;
	}

	format.wBitsPerSample = bitdepth;
	format.nChannels = channels;
	format.nBlockAlign = channels*bitdepth/8;
	format.nSamplesPerSec = samplerate;
	format.nAvgBytesPerSec = format.nBlockAlign*samplerate;
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.cbSize = 0;

	memset(&bufferDesc,0,sizeof(bufferDesc));
	bufferDesc.dwSize = sizeof(bufferDesc);
	bufferDesc.dwFlags = DSBCAPS_STICKYFOCUS;
	bufferDesc.dwBufferBytes = BUFFER_SIZE*BUFFER_STEPS;
	bufferDesc.dwReserved = 0;
	bufferDesc.lpwfxFormat = &format;
//	bufferDesc.guid3DAlgorithm = GUID_NULL;

	if (ds->CreateSoundBuffer(&bufferDesc,&soundBuffer,NULL)!=DS_OK) {
		printf("Can't create sound buffer\n");
		return 0;
	}

	return 1;
}

void start_play(read_func rf, void *param) {
	unsigned int n, i;
	int play = 1;
	DWORD playCursor;
	LPVOID ptr1, ptr2;
	DWORD size1, size2;
	unsigned char *ptr;
	unsigned int endPos = 0;
	int extraWait = 0;

	soundBuffer->Lock(0,0,&ptr1,&size1,&ptr2,&size2,DSBLOCK_ENTIREBUFFER);
	memset(ptr1,0,BUFFER_SIZE*BUFFER_STEPS);
	soundBuffer->Unlock(ptr1,BUFFER_SIZE*BUFFER_STEPS,ptr2,0);

	soundBuffer->Play(0,0,DSBPLAY_LOOPING);

	i = 0;
	stop = 0;
	playing = 1;
	while (play && !stop) {
		while (1) {
			soundBuffer->GetCurrentPosition(&playCursor,NULL);
			if (playCursor < i*BUFFER_SIZE || playCursor > (i+1)*BUFFER_SIZE) break;
			Sleep(SLEEP_LENGTH);
		}
		soundBuffer->Lock(i*BUFFER_SIZE,0,&ptr1,&size1,&ptr2,&size2,DSBLOCK_ENTIREBUFFER);
		ptr = (unsigned char*) ptr1;
		n = rf(param,ptr,BUFFER_SIZE);
		if (n<=0) {
			endPos = i*BUFFER_SIZE;
			extraWait = playCursor>=endPos;
			play = 0;
			soundBuffer->Unlock(ptr1,0,ptr2,0);
			break;
		}
		if (n<BUFFER_SIZE)
			memset(ptr+n,0,BUFFER_SIZE-n);
		soundBuffer->Unlock(ptr1,BUFFER_SIZE,ptr2,0);
		i++;
		if (i==BUFFER_STEPS) i = 0;
	}
	if (stop) {
		soundBuffer->Stop();
		return;
	}
	// Wait for play cursor to reach end of buffer, so we can check for playCursor >= endPos
	if (extraWait) {
		play = 1;
		// Because playCursor is initially further in the buffer than our writepos,
		// playCursor has restarted when the writepos reaches the end and restarts
		while (i < BUFFER_STEPS && play) {
			while (1) {
				soundBuffer->GetCurrentPosition(&playCursor,NULL);
				// If playCursor is less than endPos, the cursor has restarted
				if (playCursor < endPos) {
					play = 0;
					break;
				}
				// No need to check for playCursor < i*BUFFER_SIZE, because if it is,
				// the cursor also should hade restarted
				if (playCursor > (i+1)*BUFFER_SIZE) break;
				Sleep(SLEEP_LENGTH);
			}
			if (!play) break;
			soundBuffer->Lock(i*BUFFER_SIZE,0,&ptr1,&size1,&ptr2,&size2,DSBLOCK_ENTIREBUFFER);
			ptr = (unsigned char*) ptr1;
			memset(ptr,0,BUFFER_SIZE);
			soundBuffer->Unlock(ptr1,BUFFER_SIZE,ptr2,0);
			i++;
		}
		if (i==BUFFER_STEPS) i = 0;
	}
	play = 1;
	// Just wait until the playCursor has passed endPos, then quit
	while (play) {
		while (1) {
			soundBuffer->GetCurrentPosition(&playCursor,NULL);
			if (playCursor >= endPos) {
				play = 0;
				break;
			}
			if (playCursor < i*BUFFER_SIZE || playCursor > (i+1)*BUFFER_SIZE) break;
			Sleep(SLEEP_LENGTH);
		}
		if (!play) break;
		soundBuffer->Lock(i*BUFFER_SIZE,0,&ptr1,&size1,&ptr2,&size2,DSBLOCK_ENTIREBUFFER);
		ptr = (unsigned char*) ptr1;
		memset(ptr,0,BUFFER_SIZE);
		soundBuffer->Unlock(ptr1,BUFFER_SIZE,ptr2,0);
		i++;
		if (i==BUFFER_STEPS) i = 0;
	}

	soundBuffer->Stop();

	playing = 0;
}

int is_playing(void) {
	return playing;
}

void stop_audio(void) {
	ds->Release();
}

void break_play(void) {
	stop = 1;
}

