#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <math.h>
#include <string.h>
#include "loadpng.h"
#include "modfile.h"
#include "modplayer.h"
#include "audio.h"
#include "thread.h"

#include "neon.h"
#include "swirl.h"
#include "growing.h"
#include "mini.h"
#include "circular.h"
#include "rotoplasma.h"
#include "invert.h"
#include "cube.h"
#include "glass.h"
#include "bumpmap.h"
#include "showtext.h"


#define WIDTH 320
#define HEIGHT 240

SDL_Surface *screen;
SDL_Surface *sdl_buffer;

unsigned int *lock() {
	if (SDL_MUSTLOCK(sdl_buffer)) SDL_LockSurface(sdl_buffer);
	return (unsigned int*) sdl_buffer->pixels;
}

void unlockFlip() {
	if (SDL_MUSTLOCK(sdl_buffer)) SDL_UnlockSurface(sdl_buffer);
	SDL_BlitSurface(sdl_buffer,NULL,screen,NULL);
	SDL_Flip(screen);
}

static int part;
static int curTime, lastTime;

static void newPattern(int orderNr) {
	part++;
	lastTime = curTime;
}

static void copyLine(unsigned int *out, unsigned int *in, int x) {
	out += x;
	in += x;
	for (int i = 0; i < HEIGHT; i++) {
		*out = *in;
		out += WIDTH;
		in += WIDTH;
	}
}

static bool quit = false;
static bool checkBreak(void) {
	SDL_Event event;
	while (SDL_PollEvent(&event)) {
		switch (event.type) {
		case SDL_KEYDOWN:
		case SDL_QUIT:
			quit = true;
			return true;
		}
	}
	return false;
}

static void wipeInRight(unsigned int *title) {
	int step, stepWidth;

	step = 8;
	stepWidth = (int) ceil((float)WIDTH/step);

	for (int i = 0; i < stepWidth+step-1; i++) {
		if (checkBreak()) break;
		unsigned int *buffer = lock();
		for (int j = 0; j < step; j++) {
			int x = ((i+1)*step-1-(step+1)*j);
			if (x>=0 && x<WIDTH)
				copyLine(buffer,title,x);
		}
		unlockFlip();
		msleep(35);
	}
}

static void drawLine(unsigned int *out, int x) {
	out += x;
	for (int i = 0; i < HEIGHT; i++) {
		*out = 0;
		out += WIDTH;
	}
}

static void wipeOutRight(void) {
	int step, stepWidth;

	step = 8;
	stepWidth = (int) ceil((float)WIDTH/step);

	for (int i = 0; i < stepWidth+step-1; i++) {
		if (checkBreak()) break;
		unsigned int *buffer = lock();
		for (int j = 0; j < step; j++) {
			int x = ((i+1)*step-1-(step+1)*j);
			if (x>=0 && x<WIDTH)
				drawLine(buffer,x);
		}
		unlockFlip();
		msleep(35);
	}
}

static void grayScale(unsigned int *ptr, int length) {
	unsigned int r,g,b;
	for (int i = 0; i < length; i++) {
		r = (*ptr&0x00FF0000)>>16;
		g = (*ptr&0x0000FF00)>>8;
		b = (*ptr&0x000000FF);
		*ptr++ = (r+g+b)/3;
	}
}

static void fadeFromBlack(unsigned int *out, unsigned int *in, int time, int duration) {
	unsigned char r,g,b;
	for (int i = 0; i < WIDTH*HEIGHT; i++) {
		r = time*((*in&0x00FF0000)>>16)/duration;
		g = time*((*in&0x0000FF00)>>8)/duration;
		b = time*(*in&0x000000FF)/duration;
		in++;
		*out++ = r<<16 | g<<8 | b;
	}
}

static void fade(unsigned int *out, unsigned int *in1, unsigned int *in2, int time, int duration) {
	unsigned char r1,g1,b1,r2,g2,b2;
	unsigned char r,g,b;
	for (int i = 0; i < WIDTH*HEIGHT; i++) {
		r1 = (*in1&0x00FF0000)>>16;
		g1 = (*in1&0x0000FF00)>>8;
		b1 = *in1&0x000000FF;
		r2 = (*in2&0x00FF0000)>>16;
		g2 = (*in2&0x0000FF00)>>8;
		b2 = *in2&0x000000FF;
		in1++;
		in2++;

		r = r1 + time*(r2-r1)/duration;
		g = g1 + time*(g2-g1)/duration;
		b = b1 + time*(b2-b1)/duration;

		*out++ = r<<16 | g<<8 | b;
	}
}

static void fadeToBlack(unsigned int *out, unsigned int *in, int time, int duration) {
	unsigned char r,g,b;
	for (int i = 0; i < WIDTH*HEIGHT; i++) {
		r = (duration-time)*((*in&0x00FF0000)>>16)/duration;
		g = (duration-time)*((*in&0x0000FF00)>>8)/duration;
		b = (duration-time)*(*in&0x000000FF)/duration;
		in++;
		*out++ = r<<16 | g<<8 | b;
	}
}

static void drop(unsigned int *in1, unsigned int *in2, int time, int duration) {
	int y = HEIGHT*time/duration;
	memcpy(in1,in2+WIDTH*HEIGHT-y*WIDTH,y*WIDTH*sizeof(int));
}




int main(int argc, char *argv[]) {
	FILE *in;

	if (SDL_Init(SDL_INIT_VIDEO)<0) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 1;
	}
	if (atexit(SDL_Quit)<0) {
		fprintf(stderr,"atexit(...)\n");
		SDL_Quit();
		return 1;
	}

	in = fopen("font.png","rb");
	if (in==NULL) {
		perror("font.png");
		return 1;
	}
	int fontWidth = 0, fontHeight = 0;
	unsigned int *font = loadPng(in,&fontWidth,&fontHeight);
	fclose(in);
	if (fontWidth!=160 || fontHeight!=112) {
		fprintf(stderr,"%s: Bad font file, %s\n",argv[0],"font.png");
		free(font);
		return 1;
	}

	in = fopen("title.png","rb");
	if (in==NULL) {
		perror("title.png");
		free(font);
		return 1;
	}
	int titleWidth = 0, titleHeight = 0;
	unsigned int *title = loadPng(in,&titleWidth,&titleHeight);
	fclose(in);
	if (titleWidth!=320 || titleHeight!=240) {
		fprintf(stderr,"%s: Bad image, %s\n",argv[0],"title.png");
		free(font);
		free(title);
		return 1;
	}

	in = fopen("bumpmap.png","rb");
	if (in==NULL) {
		perror("bumpmap.png");
		free(font);
		free(title);
		return 1;
	}
	int bumpMapWidth = 0, bumpMapHeight = 0;
	unsigned int *bumpMap = loadPng(in,&bumpMapWidth,&bumpMapHeight);
	fclose(in);
	if (bumpMapWidth!=321 || bumpMapHeight!=241) {
		fprintf(stderr,"%s: Bad image, %s\n",argv[0],"bumpmap.png");
		free(font);
		free(title);
		free(bumpMap);
		return 1;
	}

	in = fopen("caffeine.mod","rb");
	if (in==NULL) {
		perror("caffeine.mod");
		free(font);
		free(title);
		free(bumpMap);
		return 1;
	}

	ModFile *mod = new ModFile(in);

	if (!init_audio()) {
		free(font);
		free(title);
		free(bumpMap);
		delete mod;
		return 1;
	}


	SDL_WM_SetCaption("Caffeine","Caffeine");
	screen = SDL_SetVideoMode(WIDTH,HEIGHT,32,SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_FULLSCREEN);
	if (screen==NULL) {
		fprintf(stderr,"Can't set video mode (%s)\n",SDL_GetError());
		return 1;
	}
	SDL_ShowCursor(SDL_DISABLE);
	SDL_Delay(750);
	sdl_buffer = SDL_CreateRGBSurface(SDL_SWSURFACE,WIDTH,HEIGHT,32,0x00FF0000,0x0000FF00,0x000000FF,0);
	if (sdl_buffer==NULL) {
		fprintf(stderr,"Can't create video buffer (%s)\n",SDL_GetError());
		return 1;
	}

	Neon *neon = NULL;
	Swirl *swirl = NULL;
	Growing *growing = NULL;
	Mini *mini = NULL;
	Circular *circular = NULL;
	RotoPlasma *rotoplasma = NULL;
	Invert *invert = NULL;
	Cube *cube = NULL;
	Glass *glass = NULL;
	BumpMap *bumpmap = NULL;
	ShowText *showtext = NULL;



#define CHECKBREAK()			\
	if (quit || checkBreak()) {	\
		delete neon;		\
		delete swirl;		\
		delete growing;		\
		delete mini;		\
		delete circular;	\
		delete rotoplasma;	\
		delete invert;		\
		delete cube;		\
		delete glass;		\
		delete bumpmap;		\
		delete showtext;	\
		free(font);		\
		free(title);		\
		free(bumpMap);		\
		delete mod;		\
		SDL_FreeSurface(sdl_buffer);	\
		return 1;		\
	}

	wipeInRight(title);
	CHECKBREAK();

	int initStartTime = SDL_GetTicks();

	grayScale(bumpMap,321*241);
	CHECKBREAK();
	neon = new Neon(WIDTH,HEIGHT);
	neon->init();
	CHECKBREAK();
	swirl = new Swirl(WIDTH,HEIGHT,1250);
	swirl->init();
	CHECKBREAK();
	growing = new Growing(WIDTH,HEIGHT);
	growing->init();
	CHECKBREAK();
	mini = new Mini(WIDTH,HEIGHT);
	mini->init();
	CHECKBREAK();
	circular = new Circular(WIDTH,HEIGHT);
	circular->init();
	CHECKBREAK();
	rotoplasma = new RotoPlasma(WIDTH,HEIGHT);
	rotoplasma->init();
	CHECKBREAK();
	invert = new Invert(WIDTH,HEIGHT);
	invert->init();
	CHECKBREAK();
	cube = new Cube(WIDTH,HEIGHT,0x00FF0000);
	cube->init();
	CHECKBREAK();
	glass = new Glass(WIDTH,HEIGHT);
	glass->init();
	CHECKBREAK();
	bumpmap = new BumpMap(WIDTH,HEIGHT);
	bumpmap->init();
	CHECKBREAK();
	showtext = new ShowText(WIDTH,HEIGHT);
	showtext->init();
	CHECKBREAK();
	showtext->setFont(font," !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]",false,16,16,10,true,0x00FF0000);
	char *text[] = {
		"Caffeine",
		"Code & Graphics\n\nwbs",
		"Music\n\nThe P",
		"Thanks to all\nmy friends.\n\nYou know who\nyou are.",
		"Greets to\neverybody at\nAssembly 01!",
		NULL
	};
	showtext->setText(text,17500);
	CHECKBREAK();

	while ((SDL_GetTicks()-initStartTime)<2500) {
		msleep(5);
		CHECKBREAK();
	}

	wipeOutRight();
	CHECKBREAK();


	part = -1;
	bool done = false;
	bool modFade = false;
	int step = -1, lastStep = 0;
	unsigned int *buffer;
	unsigned int *buffer2 = (unsigned int*) malloc(sizeof(int)*WIDTH*HEIGHT);
	int frames = 0;
//	int lastStepTime = 0;


	ModPlayer *player = new ModPlayer(mod,channels,samplerate,bitdepth,sig,littleendian);
	player->addPatternCallback(newPattern);
	start_thread(player);

	int startTime = SDL_GetTicks();
	SDL_Event event;
	while (!done) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_KEYDOWN:
			case SDL_QUIT:
				done = true;
				break;
			}
		}
		curTime = SDL_GetTicks()-startTime;
		if (part<0) continue;
		frames++;

		buffer = lock();

		if (part<2) {
			if (step<=0 && curTime-lastTime<500) {
				step = 0;
				neon->draw(buffer,curTime);
				fadeFromBlack(buffer,buffer,curTime-lastTime,500);
			} else {
				step = 1;
				neon->draw(buffer,curTime);
			}
		} else if (part<4) {
			if (step<=2 && curTime-lastTime<750) {
				step = 2;
				neon->draw(buffer,curTime);
				swirl->draw(buffer2,curTime);
				growing->draw(buffer,buffer,buffer2,curTime-lastTime,750);
			} else {
				step = 3;
				swirl->draw(buffer,curTime);
			}
		} else if (part<5) {
			if (step<=4 && curTime-lastTime<500) {
				step = 4;
				swirl->draw(buffer,curTime);
				memcpy(buffer2,buffer,WIDTH*HEIGHT*sizeof(int));
				mini->draw(buffer2);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 5;
				swirl->draw(buffer,curTime);
				mini->draw(buffer);
			}
		} else if (part<7) {
			if (step<=6 && curTime-lastTime<500) {
				step = 6;
				swirl->draw(buffer,curTime);
				mini->draw(buffer);
				circular->draw(buffer2,curTime);
				drop(buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 7;
				circular->draw(buffer,curTime);
			}
		} else if (part<10) {
			if (step<=8 && curTime-lastTime<500) {
				step = 8;
				circular->draw(buffer,curTime);
				rotoplasma->draw(buffer2,curTime);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 9;
				rotoplasma->draw(buffer,curTime);
			}
		} else if (part<12) {
			if (step<=10 && curTime-lastTime<500) {
				step = 10;
				rotoplasma->draw(buffer,curTime);
				neon->draw(buffer2,curTime);
				invert->draw(buffer2,curTime);
				mini->draw(buffer2);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				neon->draw(buffer,curTime);
				invert->draw(buffer,curTime);
				mini->draw(buffer);
				step = 11;
			}
		} else if (part<14) {
			if (step<=12 && curTime-lastTime<500) {
				step = 12;
				neon->draw(buffer,curTime);
				circular->draw(buffer2,curTime);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
				invert->draw(buffer,curTime);
				mini->draw(buffer);
			} else {
				step = 13;
				circular->draw(buffer,curTime);
				invert->draw(buffer,curTime);
				mini->draw(buffer);
			}
		} else if (part<15) {
			if (step<=14 && curTime-lastTime<500) {
				step = 14;
				circular->draw(buffer,curTime);
				invert->draw(buffer,curTime);
				mini->draw(buffer);
				neon->draw(buffer2,curTime);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 15;
				neon->draw(buffer,curTime);
			}
		} else if (part<16) {
			if (step<=16 && curTime-lastTime<500) {
				step = 16;
				neon->draw(buffer,curTime);
				memcpy(buffer2,buffer,WIDTH*HEIGHT*sizeof(int));
				cube->draw(buffer2,curTime,false);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 17;
				neon->draw(buffer,curTime);
				cube->draw(buffer,curTime,false);
			}
		} else if (part<18) {
			if (step<=18 && curTime-lastTime<500) {
				step = 18;
				neon->draw(buffer,curTime);
				glass->draw(buffer2,curTime,buffer,bumpMap);
				cube->draw(buffer,curTime,false);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 19;
				neon->draw(buffer,curTime);
				glass->draw(buffer,curTime,buffer,bumpMap);
			}
		} else if (part<20) {
			if (step<=20 && curTime-lastTime<500) {
				step = 20;
				neon->draw(buffer,curTime);
				bumpmap->draw(buffer2,curTime,buffer,bumpMap);
				glass->draw(buffer,curTime,buffer,bumpMap);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 21;
				neon->draw(buffer,curTime);
				bumpmap->draw(buffer,curTime,buffer,bumpMap);
			}
		} else if (part<25) {
			if (step<=22 && curTime-lastTime<500) {
				step = 22;
				neon->draw(buffer2,curTime);
				bumpmap->draw(buffer,curTime,buffer2,bumpMap);
				fade(buffer,buffer,buffer2,curTime-lastTime,500);
			} else {
				step = 23;
				neon->draw(buffer,curTime);
				showtext->draw(buffer,curTime);
			}
		} else if (part<26) {
			if (step<=24 && curTime-lastTime<500) {
				step = 24;
				if (!modFade) {
					player->fadeOut(500);
					modFade = true;
				}
				neon->draw(buffer,curTime);
				fadeToBlack(buffer,buffer,curTime-lastTime,500);
			} else {
				memset(buffer,0,WIDTH*HEIGHT*sizeof(int));
				done = true;
			}
		} else {
			done = true;
		}

		if (step!=lastStep) {
//			if (frames>1)
//				printf("%d: %d fps\n",lastStep,1000*(frames-1)/(curTime-lastStepTime));
			frames = 1;
			lastStep = step;
//			lastStepTime = curTime;
		}


		unlockFlip();
	}

	break_play();

	join_thread();


	delete player;
	free(buffer2);

	delete neon;
	delete swirl;
	delete growing;
	delete mini;
	delete circular;
	delete rotoplasma;
	delete invert;
	delete cube;
	delete glass;
	delete bumpmap;
	delete showtext;

	free(font);
	free(title);
	free(bumpMap);
	delete mod;

	SDL_FreeSurface(sdl_buffer);

	return 0;
}

