#include <stdlib.h>
#include <math.h>
#include "neon.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

Neon::Neon(int w, int h) {
	width = w;
	height = h;
	sintable = NULL;
	disttable = NULL;
}

Neon::~Neon(void) {
	free(sintable);
	free(disttable);
}

void Neon::init(void) {
	int i;
	xvar = (512-width)/2;
	yvar = (512-height)/2;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (unsigned char) (128+127*sin(2*M_PI*i/512));

	disttable = (unsigned int*) malloc(sizeof(int)*512*512);
	unsigned int *ptr = disttable;
	for (int y = 0; y < 512; y++)
		for (int x = 0; x < 512; x++)
			*ptr++ = (int)(5*hypot(x-256,y-256));

	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Neon::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	int i;

	for (i = 0; i < 128; i++)
		*ptr++ = 128+i;
	for (i = 128; i < 192; i++)
		*ptr++ = ((i-128)<<2)<<8 | 255;
	for (i = 192; i < 256; i++)
		*ptr++ = ((i-192)<<2)<<16 | 255<<8 | 255;

	return pal;
}

void Neon::draw(unsigned int *buffer, int time) {
	int offsetX1 = (int) (xvar+xvar*sin((double)(3950+time)/900));
	int offsetY1 = (int) (yvar+yvar*cos((double)(2650+time)/750));
	int offsetX2 = (int) (xvar+xvar*sin((double)(4900+time)/850));
	int offsetY2 = (int) (yvar+yvar*cos((double)(1250+time)/1000));
	int offsetX3 = (int) (xvar+xvar*sin((double)(250+time)/750));
	int offsetY3 = (int) (yvar+yvar*cos((double)(2550+time)/950));
	int offsetX4 = (int) (xvar+xvar*sin((double)(2200+time)/1200));
	int offsetY4 = (int) (yvar+yvar*cos((double)(500+time)/1550));

	unsigned int *ptr1, *ptr2, *ptr3, *ptr4;

	for (int y = 0; y < height; y++) {
		ptr1 = disttable + ((y+offsetY1)<<9) + offsetX1;
		ptr2 = disttable + ((y+offsetY2)<<9) + offsetX2;
		ptr3 = disttable + ((y+offsetY3)<<9) + offsetX3;
		ptr4 = disttable + ((y+offsetY4)<<9) + offsetX4;
		for (int x = 0; x < width; x++)
			*buffer++ = sintable[(*ptr1++ + *ptr2++ + *ptr3++ - *ptr4++)&0x01FF];
	}
}


