#include <stdio.h>
#include <string.h>
#include <SDL.h>
#include "audio.h"

static int stop;
static int playing = 0;

int channels = 2;
int samplerate = 44100;
int bitdepth = 16;
int sig = 1;
int littleendian = 1;

struct play_data {
	read_func rf;
	void *ptr;
};
static struct play_data pd;
static int quit;

#define QUIT_NONE 0
#define QUIT_SUB 1
#define QUIT_ALL 2


void callback(void *userdata, Uint8 *stream, int len) {
	struct play_data *pd = (struct play_data*) userdata;
	int n;
	n = pd->rf(pd->ptr,stream,len);
#ifdef NOTHREAD
	if (stop) n = 0;
#endif
	if (n==0) {
		memset(stream,0,len);
		SDL_PauseAudio(1);
	} else if (n<len) {
		memset(stream+n,0,len-n);
	}
}

int init_audio(void) {
	SDL_AudioSpec desired, obtained;
	Uint32 inited;

	if (((inited=SDL_WasInit(SDL_INIT_EVERYTHING))&SDL_INIT_AUDIO)==0) {
		if (inited>0) {
			if (SDL_InitSubSystem(SDL_INIT_AUDIO)<0) {
				fprintf(stderr,"%s\n",SDL_GetError());
				return 0;
			}
			quit = QUIT_SUB;
		} else {
			if (SDL_Init(SDL_INIT_AUDIO)<0) {
				fprintf(stderr,"%s\n",SDL_GetError());
				return 0;
			}
			quit = QUIT_ALL;
		}
	} else {
		quit = QUIT_NONE;
	}
	desired.freq = samplerate;
	desired.format = AUDIO_S16LSB;
	desired.channels = channels;
	desired.samples = 2048;
	desired.callback = callback;
	desired.userdata = (void*) &pd;
	if (SDL_OpenAudio(&desired,&obtained)<0) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 0;
	}

	samplerate = obtained.freq;
	channels = obtained.channels;
	switch (obtained.format) {
	case AUDIO_U8:
		bitdepth = 8;
		sig = 0;
		break;
	case AUDIO_S8:
		bitdepth = 8;
		sig = 1;
		break;
	case AUDIO_U16LSB:
		bitdepth = 16;
		sig = 0;
		littleendian = 1;
		break;
	case AUDIO_S16LSB:
		bitdepth = 16;
		sig = 1;
		littleendian = 1;
		break;
	case AUDIO_U16MSB:
		bitdepth = 16;
		sig = 0;
		littleendian = 0;
		break;
	case AUDIO_S16MSB:
		bitdepth = 16;
		sig = 1;
		littleendian = 0;
		break;
	default:
		fprintf(stderr,"Unable to set audio format\n");
		return 0;
	}

	return 1;
}

void stop_audio(void) {
	SDL_CloseAudio();

	switch (quit) {
	case QUIT_NONE:
		break;
	case QUIT_SUB:
		SDL_QuitSubSystem(SDL_INIT_AUDIO);
		break;
	case QUIT_ALL:
		SDL_Quit();
		break;
	}
}

void start_play(read_func rf, void *ptr) {
	pd.rf = rf;
	pd.ptr = ptr;
	stop = 0;
	playing = 1;
	SDL_PauseAudio(0);
#ifndef NOTHREAD
	while (SDL_GetAudioStatus()==SDL_AUDIO_PLAYING && !stop) SDL_Delay(10);
	if (stop) SDL_PauseAudio(1);
	playing = 0;
	stop = 1;
#endif
}

int is_playing(void) {
#ifndef NOTHREAD
	return playing;
#else
	return SDL_GetAudioStatus()==SDL_AUDIO_PLAYING && !stop;
#endif
}

void break_play(void) {
	stop = 1;
}

