#include <stdlib.h>
#include <math.h>
#include "swirl.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

Swirl::Swirl(int w, int h, int p) {
	width = w;
	height = h;
	period = p;
	sintable = NULL;
	angletable = NULL;
	disttable = NULL;
}

Swirl::~Swirl(void) {
	free(sintable);
	free(angletable);
	free(disttable);
}

void Swirl::init(void) {
	int i,x,y;
	trigFact = (float)M_PI/period;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (int)(128+127*sin(2*M_PI*i/512));

	int middleX = width/2, middleY = height/2;

	angletable = (unsigned int*) malloc(sizeof(int)*width*height);
	unsigned int *ptr = angletable;
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
			*ptr++ = (int)((1<<16)*512*atan2(y-middleY,x-middleX)/(2*M_PI));
	disttable = (unsigned int*) malloc(sizeof(int)*width*height);
	ptr = disttable;
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
			*ptr++ = (int)((1<<16)*hypot(x-middleX,y-middleY)/10);

	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Swirl::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	int i;

	for (i = 0; i < 128; i++)
		*ptr++ = (128-i)<<1;
	for (i = 128; i < 256; i++)
		*ptr++ = ((i-128)<<1)<<8;

	return pal;
}

void Swirl::draw(unsigned int *buffer, int time, int multiplier, bool direction, int turnTime) {
	int curvyness = (int) ((1<<8)*50*sin(trigFact*time));
	int spin;
	if (!direction)
		spin = (int) (time/1.0);
	else
		spin = (int) ((2*turnTime-time)/1.0);

	unsigned int *angleptr = angletable;
	unsigned int *distptr = disttable;
	for (int y = 0; y < height; y++)
		for (int x = 0; x < width; x++)
			*buffer++ = sintable[(spin+((multiplier*(*angleptr++)+curvyness*((*distptr++)>>8))>>16))&0x01FF];
}


